import os
import unittest

import numpy as np

import rubin_scheduler.utils as utils
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler import sim_runner
from rubin_scheduler.scheduler.basis_functions import SunAltLimitBasisFunction
from rubin_scheduler.scheduler.example import example_scheduler, simple_greedy_survey, simple_pairs_survey
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.schedulers import CoreScheduler

SAMPLE_BIG_DATA_FILE = os.path.join(get_data_dir(), "scheduler/dust_maps/dust_nside_32.npz")


class ModelObservatoryWindy(ModelObservatory):
    """Have the model observatory always have a strong wind from the north"""

    def return_conditions(self):
        """
        Returns
        -------
        rubin_scheduler.scheduler.features.conditions object
        """
        _conditions = super().return_conditions()

        # Always have a strong wind from the north
        wind_speed = 40.0
        wind_direction = 0.0
        self.conditions.wind_speed = wind_speed
        self.conditions.wind_direction = wind_direction

        return self.conditions


class TestExample(unittest.TestCase):
    @unittest.skipUnless(os.path.isfile(SAMPLE_BIG_DATA_FILE), "Test data not available.")
    def test_example(self):
        """Try out the example scheduler."""
        mjd_start = utils.survey_start_mjd()
        nside = 32
        survey_length = 4.0  # days
        scheduler = example_scheduler(nside=nside, mjd_start=mjd_start)
        observatory = ModelObservatory(nside=nside, mjd_start=mjd_start)
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, sim_duration=survey_length, filename=None
        )
        # Check that greedy observed some
        assert "greedy" in observations["scheduler_note"]
        # check some long pairs got observed
        assert np.any(["pair_33" in obs for obs in observations["scheduler_note"]])
        # Make sure lots of observations executed
        assert observations.size > 1000
        # Make sure nothing tried to look through the earth
        assert np.min(observations["alt"]) > 0
        # Make sure a DDF executed
        assert np.any(["DD" in note for note in observations["scheduler_note"]])

    @unittest.skipUnless(os.path.isfile(SAMPLE_BIG_DATA_FILE), "Test data not available.")
    def test_wind(self):
        """
        Test that a wind mask prevent things from being executed in
        the wrong spot
        """
        mjd_start = utils.survey_start_mjd()
        nside = 32
        survey_length = 2.0  # days

        surveys = [simple_greedy_survey(filtername=f) for f in "gri"]

        scheduler = CoreScheduler(surveys, nside=nside)
        observatory = ModelObservatoryWindy(
            nside=nside, mjd_start=mjd_start, downtimes="ideal", cloud_data="ideal"
        )
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, sim_duration=survey_length, filename=None
        )

        # Make sure lots of observations executed, but allow short night
        # if survey_start changes
        assert observations.size > 700 * survey_length
        # Make sure nothing tried to look through the earth
        assert np.min(observations["alt"]) > 0

        # Make sure nothing executed in the strong wind
        assert np.min(np.degrees(observations["az"])) > 30.0
        assert np.max(np.degrees(observations["az"])) < (360.0 - 30.0)

    @unittest.skipUnless(os.path.isfile(SAMPLE_BIG_DATA_FILE), "Test data not available.")
    def test_nside(self):
        """
        test running at higher nside
        """
        mjd_start = utils.survey_start_mjd()
        nside = 64
        survey_length = 2.0  # days

        # Add an avoidance of twilight+ for the pairs surveys -
        # this ensures greedy survey will have some time to operate
        pairs_surveys = [
            simple_pairs_survey(filtername="g", filtername2="r", nside=nside),
            simple_pairs_survey(filtername="i", filtername2="z", nside=nside),
        ]
        for survey in pairs_surveys:
            survey.basis_functions.append(SunAltLimitBasisFunction(alt_limit=-22))
            survey.basis_weights.append(0)
        greedy_surveys = [
            simple_greedy_survey(filtername="z", nside=nside),
        ]

        scheduler = CoreScheduler([pairs_surveys, greedy_surveys], nside=nside)
        observatory = ModelObservatory(nside=nside, mjd_start=mjd_start)
        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, sim_duration=survey_length, filename=None
        )

        # Make sure some blobs executed
        assert "simple pair 30, iz, a" in observations["scheduler_note"]
        assert "simple pair 30, iz, b" in observations["scheduler_note"]
        # Make sure some greedy executed
        assert "simple greedy z" in observations["scheduler_note"]
        # Make sure lots of observations executed
        assert observations.size > 800
        # Make sure nothing tried to look through the earth
        assert np.min(observations["alt"]) > 0


if __name__ == "__main__":
    unittest.main()



__all__ = ("get_data_dir", "data_versions")

import os


def get_data_dir():
    """Get the location of the rubin_sim data directory.

    Returns
    -------
    data_dir : `str`
        Path to the rubin_sim data directory.
    """
    # See if there is an environment variable with the path
    data_dir = os.getenv("RUBIN_SIM_DATA_DIR")

    # Set the root data directory
    if data_dir is None:
        data_dir = os.path.join(os.getenv("HOME"), "rubin_sim_data")
    return data_dir


def data_versions():
    """Get the dictionary of source filenames in the rubin_sim data directory.

    Returns
    -------
    result : `dict`
        Data directory filenames dictionary with keys/values:
        "name" - Data bucket name (`str`).
        "version" - Versioned file name (`str`).
    """
    data_dir = get_data_dir()
    result = None
    version_file = os.path.join(data_dir, "versions.txt")
    if os.path.isfile(version_file):
        with open(version_file) as f:
            content = f.readlines()
        content = [x.strip() for x in content]
        result = {}
        for line in content:
            ack = line.split(",")
            result[ack[0]] = ack[1]

    return result



__all__ = ("FieldSurvey",)

import copy
import warnings
from functools import cached_property

import numpy as np

from rubin_scheduler.utils import ra_dec2_hpid

from ..features import LastObservation, NObsCount
from ..utils import ObservationArray
from . import BaseSurvey


class FieldSurvey(BaseSurvey):
    """A survey class for running field surveys.

    Parameters
    ----------
    basis_functions : `list` [`rubin_scheduler.scheduler.basis_function`]
        List of basis_function objects
    detailers : `list` [`rubin_scheduler.scheduler.detailer`] objects
        The detailers to apply to the list of observations.
    RA : `float`
        The RA of the field (degrees)
    dec : `float`
        The dec of the field to observe (degrees)
    sequence : `list` [`str`]
        The sequence of observations to take. (specify which filters to use).
    nvisits : `dict` {`str`: `int`}
        Dictionary of the number of visits in each filter.
        Default of None will use a backup sequence of 20 visits per filter.
        Must contain all filters in sequence.
    exptime : `dict` {`str`: `float`}
        Dictionary of the exposure time for visits in each filter.
        Default of None will use a backup sequence of 38s in u, and
        29.2s in all other bands. Must contain all filters in sequence.
    nexp : dict` {`str`: `int`}
        Dictionary of the number of exposures per visit in each filter.
        Default of None will use a backup sequence of 1 exposure per visit
        in u band, 2 in all other bands. Must contain all filters in sequence.
    ignore_obs : `list` [`str`] or None
        Ignore observations with this string in the `scheduler_note`.
        Will ignore observations which match subsets of the string, as well as
        the entire string. Ignoring 'mysurvey23' will also ignore 'mysurvey2'.
    accept_obs : `list` [`str`] or None
        If match_obs is set, then ONLY observations which match these
        strings in the `scheduler_note` will be counted for the survey.
        A complete match must occur; substrings will not match.
        (for obs_array too??)
    survey_name : `str` or None.
        The name to give this survey, for debugging and visualization purposes.
        Also propagated to the 'target_name' in the observation.
        The default None will construct a name based on the
        RA/Dec of the field.
    scheduler_note : `str` or None
        The value to include in the scheduler note.
        The scheduler note is for internal, scheduler, use for the purposes of
        identifying observations to ignore or include for a survey or feature.
    readtime : `float`
        Readout time for computing approximate time of observing
        the sequence. (seconds)
    filter_change_time : `float`
        Filter change time, on average. Used for computing approximate
        time for the observing sequence. (seconds)
    nside : `float` or None
        Nside for computing survey basis functions and maps.
        The default of None will use rubin_scheduler.utils.set_default_nside().
    flush_pad : `float`
        How long to hold observations in the queue after they
        were expected to be completed (minutes).
    reward_value : `float`
        An unused kwarg, provided for backward compatibility.
    """

    def __init__(
        self,
        basis_functions,
        RA,
        dec,
        sequence="ugrizy",
        nvisits=None,
        exptimes=None,
        nexps=None,
        ignore_obs=None,
        accept_obs=None,
        survey_name=None,
        target_name=None,
        science_program=None,
        observation_reason=None,
        scheduler_note=None,
        readtime=2.4,
        filter_change_time=120.0,
        nside=None,
        flush_pad=30.0,
        detailers=None,
        reward_value=None,
        nexp=None,
    ):
        default_nvisits = {"u": 20, "g": 20, "r": 20, "i": 20, "z": 20, "y": 20}
        default_exptimes = {"u": 38, "g": 29.2, "r": 29.2, "i": 29.2, "z": 29.2, "y": 29.2}
        default_nexps = {"u": 1, "g": 2, "r": 2, "i": 2, "z": 2, "y": 2}

        # Deprecated kwarg messages
        if reward_value is not None:
            warnings.warn("reward_value has been unused and will be deprecated.")
        if nexp is not None:
            if nexps is not None:
                warnings.warn(
                    "Use one of `nexp` or `nexps`: `nexps` will be "
                    "supported going forward and will override."
                )
            if nexps is None:
                warnings.warn("Please use `nexps` in the future. " "Will adapt from `nexp` presently.")
                nexps = nexp

        self.ra = np.radians(RA)
        self.ra_hours = RA / 360.0 * 24.0
        self.dec = np.radians(dec)
        self.ra_deg, self.dec_deg = RA, dec

        self.survey_name = survey_name
        if self.survey_name is None:
            self._generate_survey_name(target_name=target_name)
        # Backfill target name if it wasn't set
        if target_name is None:
            target_name = self.survey_name

        super().__init__(
            nside=nside,
            basis_functions=basis_functions,
            detailers=detailers,
            ignore_obs=ignore_obs,
            survey_name=self.survey_name,
            target_name=target_name,
            science_program=science_program,
            observation_reason=observation_reason,
        )
        self.accept_obs = accept_obs
        if isinstance(self.accept_obs, str):
            self.accept_obs = [self.accept_obs]
        self.indx = ra_dec2_hpid(self.nside, self.ra_deg, self.dec_deg)

        # Set all basis function equal.
        self.basis_weights = np.ones(len(basis_functions)) / len(basis_functions)

        self.flush_pad = flush_pad / 60.0 / 24.0  # To days
        self.filter_sequence = []

        self.scheduler_note = scheduler_note
        if self.scheduler_note is None:
            self.scheduler_note = self.survey_name

        # This sets up what a requested "observation" looks like.
        # For sequences, each 'observation' is more than one exposure.
        # When generating actual observations, filters which are not available
        # are not included in the requested sequence.
        if nvisits is None:
            nvisits = default_nvisits
        if exptimes is None:
            exptimes = default_exptimes
        if nexps is None:
            nexps = default_nexps

        # Do a little shuffling if this was not configured quite as expected
        if isinstance(sequence, str):
            if isinstance(nvisits, (float, int)):
                nvisits = dict([(filtername, nvisits) for filtername in sequence])
            if isinstance(exptimes, (float, int)):
                exptimes = dict([(filtername, exptimes) for filtername in sequence])
            if isinstance(nexps, (float, int)):
                nexps = dict([(filtername, nexps) for filtername in sequence])

        if isinstance(sequence, str):
            self.observations = []
            for filtername in sequence:
                for j in range(nvisits[filtername]):
                    obs = ObservationArray()
                    obs["filter"] = filtername
                    obs["exptime"] = exptimes[filtername]
                    obs["RA"] = self.ra
                    obs["dec"] = self.dec
                    obs["nexp"] = nexps[filtername]
                    obs["scheduler_note"] = self.scheduler_note
                    self.observations.append(obs)
        else:
            self.observations = sequence

        # Let's just make this an array for ease of use
        self.observations = np.concatenate(self.observations)
        order = np.argsort(self.observations["filter"])
        self.observations = self.observations[order]

        n_filter_change = np.size(np.unique(self.observations["filter"]))

        # Make an estimate of how long a sequence will take.
        # Assumes no major rotational or spatial
        # dithering slowing things down.
        # Does not account for unavailable filters.
        self.approx_time = (
            np.sum(self.observations["exptime"] + readtime * self.observations["nexp"])
            + filter_change_time * n_filter_change
        )
        # convert to days, for internal approximation in timestep sizes
        self.approx_time /= 3600.0 / 24.0
        # This is the only index in the healpix arrays that will be considered
        self.indx = ra_dec2_hpid(self.nside, self.ra_deg, self.dec_deg)

        # Tucking this here so we can look at how many observations
        # recorded for this field and what was the last one.
        self.extra_features["ObsRecorded"] = NObsCount()
        self.extra_features["LastObs"] = LastObservation()

    def _generate_survey_name(self, target_name=None):
        if target_name is not None:
            self.survey_name = target_name
        else:
            self.survey_name = f"Field {self.ra_deg :.2f} {self.dec_deg :.2f}"

    @cached_property
    def roi_hpid(self):
        hpid = ra_dec2_hpid(self.nside, np.degrees(self.ra), np.degrees(self.dec))
        return hpid

    def check_continue(self, observation, conditions):
        # feasibility basis functions?
        """
        This method enables external calls to check if a given
        observations that belongs to this survey is
        feasible or not. This is called once a sequence has
        started to make sure it can continue.

        XXX--TODO:  Need to decide if we want to develop check_continue,
        or instead hold the sequence in the survey, and be able to check
        it that way.
        (note that this may depend a lot on how the SchedulerCSC works)
        """
        return True

    def add_observation(self, observation, **kwargs):
        """Add observation one at a time."""
        # Check each possible ignore string
        checks = [io not in str(observation["scheduler_note"]) for io in self.ignore_obs]
        passed_ignore = all(checks)
        passed_accept = True
        if passed_ignore and self.accept_obs is not None:
            # Check if this observation matches any accept string.
            checks = [io == str(observation["scheduler_note"][0]) for io in self.accept_obs]
            passed_accept = any(checks)
        # I think here I have to assume observation is an
        # array and not a dict.
        if passed_ignore and passed_accept:
            for feature in self.extra_features:
                self.extra_features[feature].add_observation(observation, **kwargs)
            for bf in self.extra_basis_functions:
                self.extra_basis_functions[bf].add_observation(observation, **kwargs)
            for bf in self.basis_functions:
                bf.add_observation(observation, **kwargs)
            for detailer in self.detailers:
                detailer.add_observation(observation, **kwargs)
            self.reward_checked = False

    def add_observations_array(self, observations_array_in, observations_hpid_in):
        """Add an array of observations rather than one at a time

        Parameters
        ----------
        observations_array_in : ObservationArray
            An array of completed observations,
            rubin_scheduler.scheduler.utils.ObservationArray
        observations_hpid_in : np.array
            Same as observations_array_in, but larger and with an
            additional column for HEALpix id. Each observation is
            listed mulitple times, once for every HEALpix it overlaps.
        """
        # Just to be sure things are sorted
        observations_array_in.sort(order="mjd")
        observations_hpid_in.sort(order="mjd")

        # Copy so we don't prune things for other survey objects
        observations_array = observations_array_in.copy()
        observations_hpid = observations_hpid_in.copy()

        for ig in self.ignore_obs:
            not_ignore = np.where(np.char.find(observations_array["scheduler_note"], ig) == -1)[0]
            observations_array = observations_array[not_ignore]

            not_ignore = np.where(np.char.find(observations_hpid["scheduler_note"], ig) == -1)[0]
            observations_hpid = observations_hpid[not_ignore]

        if self.accept_obs is not None:
            accept_indx = []
            accept_hp_indx = []
            for acc in self.accept_obs:
                accept_indx.append(np.where(observations_array["scheduler_note"] == acc)[0])
                accept_hp_indx.append(np.where(observations_hpid["scheduler_note"] == acc)[0])
            accept = np.concatenate(accept_indx)
            accept_hp = np.concatenate(accept_hp_indx)
            observations_array = observations_array[accept]
            observations_hpid = observations_hpid[accept_hp]

        for feature in self.extra_features:
            self.extra_features[feature].add_observations_array(observations_array, observations_hpid)
        for bf in self.extra_basis_functions:
            self.extra_basis_functions[bf].add_observations_array(observations_array, observations_hpid)
        for bf in self.basis_functions:
            bf.add_observations_array(observations_array, observations_hpid)
        for detailer in self.detailers:
            detailer.add_observations_array(observations_array, observations_hpid)
        self.reward_checked = False

    def calc_reward_function(self, conditions):
        # only calculates reward at the index for the RA/Dec of the field
        self.reward_checked = True
        if self._check_feasibility(conditions):
            self.reward = 0
            for bf, weight in zip(self.basis_functions, self.basis_weights):
                basis_value = bf(conditions, indx=self.indx)
                self.reward += basis_value * weight

            if not np.isscalar(self.reward):
                self.reward = np.sum(self.reward[self.indx])

                if np.any(np.isinf(self.reward)):
                    self.reward = np.inf
        else:
            # If not feasible, negative infinity reward
            self.reward = -np.inf

        return self.reward

    def generate_observations_rough(self, conditions):
        result = []
        if self._check_feasibility(conditions):
            result = copy.deepcopy(self.observations)

            # Set the flush_by
            result["flush_by_mjd"] = conditions.mjd + self.approx_time + self.flush_pad

            # remove filters that are not mounted
            mask = np.isin(result["filter"], conditions.mounted_filters)
            result = result[mask]
            # Put current loaded filter first
            ind1 = np.where(result["filter"] == conditions.current_filter)[0]
            ind2 = np.where(result["filter"] != conditions.current_filter)[0]
            result = result[ind1.tolist() + (ind2.tolist())]

            # convert to list of array.
            final_result = [
                row.reshape(
                    1,
                )
                for row in result
            ]
            result = final_result

        return result

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} survey_name='{self.survey_name}'"
            f", RA={self.ra}, dec={self.dec} at {hex(id(self))}>"
        )



__all__ = ("SpecMap", "default_spec_map")

import os
import re


class SpecMap:
    subdir_map = {
        "(^km)|(^kp)": "starSED/kurucz",
        "(^bergeron)": "starSED/wDs",
        "(^burrows)|(^(m|L|l)[0-9])": "starSED/old_mlt",
        "(^lte)": "starSED/mlt",
        "^(Exp|Inst|Burst|Const)": "galaxySED",
    }

    def __init__(self, file_dict=None, dir_dict=None):
        """
        Parameters
        ----------
        file_dict : `Unknown`
            is a dict mapping the names of files to their
            relative paths, one-to-one, e.g.

            file_dict = {'A.dat':'ssmSED/A.dat.gz',
                        'Sa.dat':'ssmSED/Sa.dat.gz'}

        Parameters
        ----------
        dir_dict : `Unknown`
            is a dict mapping forms of names of file to their
            sub directories using regular expressions, e.g.

            dir_dict = {'(^km)|(^kp)':'starSED/kurucz',
                       '(^bergeron)':'starSED/wDs'}

        which maps files begining in either 'km' or 'kp' to files with
        the same names in the directory starSED/kurucz and maps files
        beginning with 'bergeron' to files with the same names in the
        directory starSED/wDs

        These dicts will take precedence over the subdir_map that is defined
        as a class member variable of the SpecMap class.
        """
        if file_dict:
            self.file_dict = file_dict
        else:
            self.file_dict = {}

        if dir_dict:
            self.dir_dict = dir_dict
        else:
            self.dir_dict = {}

    def __setitem__(self, key, val):
        self.file_dict[key] = val

    def __getitem__(self, item):
        item = item.strip()

        if item in self.file_dict:
            return self.file_dict[item]

        for key, val in sorted(self.dir_dict.items()):
            if re.match(key, item):
                full_name = item if item.endswith(".gz") else item + ".gz"
                return os.path.join(val, full_name)

        for key, val in sorted(self.subdir_map.items()):
            if re.match(key, item):
                full_name = item if item.endswith(".gz") else item + ".gz"
                return os.path.join(val, full_name)

        raise KeyError("No path found for spectrum name: %s" % (item))

    def __contains__(self, item):
        """
        Returns True if there is a map for 'item'; False if not.

        This exists primarily so that phoSim input catalog classes
        can identify columns which have no sedFilePath and then remove
        them when writing the catalog.
        """
        try:
            self.__getitem__(item)
            return True
        except KeyError:
            return False


default_spec_map = SpecMap(
    file_dict={
        "A.dat": "ssmSED/A.dat.gz",
        "Sa.dat": "ssmSED/Sa.dat.gz",
        "O.dat": "ssmSED/O.dat.gz",
        "harris_V.dat": "ssmSED/harris_V.dat.gz",
        "Cg.dat": "ssmSED/Cg.dat.gz",
        "Sv.dat": "ssmSED/Sv.dat.gz",
        "X.dat": "ssmSED/X.dat.gz",
        "K.dat": "ssmSED/K.dat.gz",
        "L.dat": "ssmSED/L.dat.gz",
        "D.dat": "ssmSED/D.dat.gz",
        "Sq.dat": "ssmSED/Sq.dat.gz",
        "C.dat": "ssmSED/C.dat.gz",
        "T.dat": "ssmSED/T.dat.gz",
        "Xk.dat": "ssmSED/Xk.dat.gz",
        "R.dat": "ssmSED/R.dat.gz",
        "Ch.dat": "ssmSED/Ch.dat.gz",
        "kurucz_sun": "ssmSED/kurucz_sun.gz",
        "Sr.dat": "ssmSED/Sr.dat.gz",
        "Cgh.dat": "ssmSED/Cgh.dat.gz",
        "Cb.dat": "ssmSED/Cb.dat.gz",
        "Xc.dat": "ssmSED/Xc.dat.gz",
        "Xe.dat": "ssmSED/Xe.dat.gz",
        "Q.dat": "ssmSED/Q.dat.gz",
        "S.dat": "ssmSED/S.dat.gz",
        "B.dat": "ssmSED/B.dat.gz",
        "V.dat": "ssmSED/V.dat.gz",
        "agn.spec": "agnSED/agn.spec.gz",
        "BD1000.dat": "starSED/gizis_SED/BD1000_interp.dat.gz",
        "BD1000e.dat": "starSED/gizis_SED/BD1000e_interp.dat.gz",
        "BD1500.dat": "starSED/gizis_SED/BD1500_interp.dat.gz",
        "BD1800.dat": "starSED/gizis_SED/BD1800_interp.dat.gz",
        "BD2000.dat": "starSED/gizis_SED/BD2000_interp.dat.gz",
        "BD325.dat": "starSED/gizis_SED/BD325_interp.dat.gz",
        "BD555.dat": "starSED/gizis_SED/BD555_interp.dat.gz",
        "PNastrom.dat": "starSED/gizis_SED/PNastrom_interp.dat.gz",
        "RedE1astrom.dat": "starSED/gizis_SED/RedE1astrom_interp.dat.gz",
        "RedE2astrom.dat": "starSED/gizis_SED/RedE2astrom_interp.dat.gz",
        "sed_flat.txt": "flatSED/sed_flat.txt.gz",
        "sed_flat_norm.txt": "flatSED/sed_flat.txt.gz",
    }
)



__all__ = [
    "get_scheduler",
    "save_scheduler",
    "add_make_scheduler_snapshot_args",
    "make_scheduler_snapshot_cli",
    "get_scheduler_instance_from_repo",
]

import argparse
import bz2
import gzip
import importlib.util
import lzma
import pickle
import sys
import types
import typing
from pathlib import Path
from tempfile import TemporaryDirectory

from git import Repo

from rubin_scheduler.scheduler.example import example_scheduler
from rubin_scheduler.scheduler.schedulers.core_scheduler import CoreScheduler


def get_scheduler_instance_from_repo(
    config_repo: str,
    config_script: str,
    config_branch: str = "main",
) -> CoreScheduler:
    """Generate a CoreScheduler according to a configuration in git.

    Parameters
    ----------
    config_repo : `str`
        The git repository with the configuration.
    config_script : `str`
        The configuration script path (relative to the repository root).
    config_branch : `str`, optional
        The branch of the repository to use, by default "main"

    Returns
    -------
    scheduler : `CoreScheduler`
        An instance of the Rubin Observatory FBS.

    Raises
    ------
    ValueError
        If the config file is invalid, or has invalid content.
    """

    with TemporaryDirectory() as local_config_repo_parent:
        repo: Repo = Repo.clone_from(config_repo, local_config_repo_parent, branch=config_branch)
        full_config_script_path: Path = Path(repo.working_dir).joinpath(config_script)
        config_module_name: str = "scheduler_config"
        config_module_spec = importlib.util.spec_from_file_location(
            config_module_name, full_config_script_path
        )
        if config_module_spec is None or config_module_spec.loader is None:
            # Make type checking happy
            raise ValueError(f"Cannot load config file {full_config_script_path}")

        config_module: types.ModuleType = importlib.util.module_from_spec(config_module_spec)
        sys.modules[config_module_name] = config_module
        config_module_spec.loader.exec_module(config_module)

    scheduler: CoreScheduler = config_module.get_scheduler()[1]
    return scheduler


def get_scheduler(
    config_repo: str | None,
    config_script: str | None,
    config_branch: str = "main",
) -> CoreScheduler:
    """Generate a CoreScheduler according to a configuration in git.

    Parameters
    ----------
    config_repo : `str`
        The git repository with the configuration.
    config_script : `str`
        The configuration script path (relative to the repository root).
    config_branch : `str`, optional
        The branch of the repository to use, by default "main"

    Returns
    -------
    scheduler : `CoreScheduler`
        An instance of the Rubin Observatory FBS.

    Raises
    ------
    ValueError
        If the config file is invalid, or has invalid content.
    """
    if config_repo is not None:
        if config_script is None:
            raise ValueError("If the config repo is set, the script must be as well.")
        scheduler = get_scheduler_instance_from_repo(
            config_repo=config_repo, config_script=config_script, config_branch=config_branch
        )
    else:
        example_scheduler_result = example_scheduler()
        if isinstance(example_scheduler_result, CoreScheduler):
            scheduler = example_scheduler_result
        else:
            # It might return a observatory, scheduler, observations tuple
            # instead.
            scheduler = example_scheduler_result[1]

    return scheduler


def save_scheduler(scheduler: CoreScheduler, file_name: str) -> None:
    """Save an instances of the scheduler in a pickle file,
    compressed according to its extension.

    Parameters
    ----------
    scheduler : `CoreScheduler`
        The scheduler to save.
    file_name : `str`
        The file in which to save the schedulers.
    """
    opener: typing.Callable = open

    if file_name.endswith(".bz2"):
        opener = bz2.open
    elif file_name.endswith(".xz"):
        opener = lzma.open
    elif file_name.endswith(".gz"):
        opener = gzip.open

    with opener(file_name, "wb") as pio:
        pickle.dump(scheduler, pio)


def add_make_scheduler_snapshot_args(parser: argparse.ArgumentParser) -> None:
    """Add arguments needed for saving a scheduler to an argument parser."""
    parser.add_argument("--scheduler_fname", type=str, help="The file in which to save the scheduler.")
    parser.add_argument(
        "--repo", type=str, default=None, help="The repository from which to load the configuration."
    )
    parser.add_argument(
        "--script", type=str, default=None, help="The path to the config script (relative to the repo root)."
    )
    parser.add_argument(
        "--branch", type=str, default="main", help="The branch of the repo from which to get the script"
    )


def make_scheduler_snapshot_cli(cli_args: list = []) -> None:
    parser = argparse.ArgumentParser(description="Create a scheduler pickle")
    add_make_scheduler_snapshot_args(parser)
    args: argparse.Namespace = parser.parse_args() if len(cli_args) == 0 else parser.parse_args(cli_args)

    scheduler: CoreScheduler = get_scheduler(args.repo, args.config, args.branch)
    save_scheduler(scheduler, args.scheduler_fname)


if __name__ == "__main__":
    make_scheduler_snapshot_cli()



__all__ = ("CoreScheduler",)

import logging
import time
from collections import OrderedDict
from copy import deepcopy
from io import StringIO

import healpy as hp
import numpy as np
import pandas as pd
from astropy.time import Time

from rubin_scheduler.scheduler.utils import (
    HpInComcamFov,
    HpInLsstFov,
    IntRounded,
    ObservationArray,
    set_default_nside,
)
from rubin_scheduler.utils import _approx_altaz2pa, _approx_ra_dec2_alt_az, _hpid2_ra_dec, rotation_converter


class CoreScheduler:
    """Core scheduler that takes requests observations,
    reports observatory status and provides completed observations.

    Parameters
    ----------
    surveys : list (or list of lists) of rubin_scheduler.scheduler.survey
        A list of surveys to consider.
        If multiple surveys return the same highest reward value,
        the survey at the earliest position in the list will be selected.
        Can also be a list of lists to make heirarchical priorities
        ('tiers').
    nside : `int`
        A HEALpix nside value.
    camera : `str`
        Which camera to use to compute the correspondence between
        visits and HEALpixels when recording observations.
        Can be 'LSST' or 'comcam'.
    rotator_limits : `list` [`float`, `float`]
        Limits for the camera rotator.
    log : `logging.Logger`
        If None, a new logger is created.
    keep_rewards : `bool`
        Flag whether to record the rewards and basis function values
        (this can be useful for schedview).
    telescope : `str`
        Which telescope model to use for rotTelPos/rotSkyPos conversions.
        Default "rubin".
    """

    def __init__(
        self,
        surveys,
        nside=None,
        camera="LSST",
        rotator_limits=[85.0, 275.0],
        log=None,
        keep_rewards=False,
        telescope="rubin",
    ):
        self.keep_rewards = keep_rewards
        # Use integer ns just to be sure there are no rounding issues.
        self.mjd_perf_counter_offset = np.int64(Time.now().mjd * 86400000000000) - time.perf_counter_ns()

        if nside is None:
            nside = set_default_nside()

        if log is None:
            self.log = logging.getLogger(type(self).__name__)
        else:
            self.log = log.getChild(type(self).__name__)

        # initialize a queue of observations to request
        self.queue = []
        # The indices of self.survey_lists that provided
        # the last addition(s) to the queue
        self.survey_index = [None, None]

        # If we have a list of survey objects, convert to list-of-lists
        if isinstance(surveys[0], list):
            self.survey_lists = surveys
        else:
            self.survey_lists = [surveys]
        self.nside = nside
        hpid = np.arange(hp.nside2npix(nside))
        self.ra_grid_rad, self.dec_grid_rad = _hpid2_ra_dec(nside, hpid)
        # Should just make camera a class that takes a pointing
        # and returns healpix indices
        if camera == "LSST":
            self.pointing2hpindx = HpInLsstFov(nside=nside)
        elif camera == "comcam":
            self.pointing2hpindx = HpInComcamFov(nside=nside)
        else:
            raise ValueError("camera %s not implamented" % camera)

        self.rc = rotation_converter(telescope=telescope)

        # keep track of how many observations get flushed from the queue
        self.flushed = 0
        self.rotator_limits = np.sort(np.radians(rotator_limits))

    def flush_queue(self):
        """Like it sounds, clear any currently queued desired observations."""
        self.queue = []
        self.survey_index = [None, None]

    def add_observations_array(self, obs):
        """Like add_observation, but for passing many observations at once.

        Assigns overlapping HEALpix IDs to each observation, then passes
        the observation array and constructed observations + healpix id
        to each survey.
        """
        obs.sort(order="mjd")
        # Generate list-of-lists for HEALPix IDs for each pointing
        lol_hpids = self.pointing2hpindx(obs["RA"], obs["dec"])
        hpids = []
        big_array_indx = []
        # Unravel the list-of-lists
        for i, indxs in enumerate(lol_hpids):
            for indx in indxs:
                hpids.append(indx)
                big_array_indx.append(i)
        hpids = np.array(hpids, dtype=[("hpid", int)])
        names = list(obs.dtype.names)
        types = [obs[name].dtype for name in names]

        names.append(hpids.dtype.names[0])
        types.append(hpids["hpid"].dtype)
        ndt = list(zip(names, types))

        obs_array_hpid = np.empty(hpids.size, dtype=ndt)
        obs_array_hpid[list(obs.dtype.names)] = obs[big_array_indx]
        obs_array_hpid[hpids.dtype.names[0]] = hpids

        for surveys in self.survey_lists:
            for survey in surveys:
                survey.add_observations_array(obs, obs_array_hpid)

    def add_observation(self, observation):
        """
        Record a completed observation and update features accordingly.

        Parameters
        ----------
        observation : dict-like
            An object that contains the relevant information about a
            completed observation
            (e.g., mjd, ra, dec, filter, rotation angle, etc)
        """

        # Catch if someone passed in a slice of an observation
        # rather than a full observation array
        if len(observation.shape) == 0:
            full_obs = ObservationArray()
            full_obs[0] = observation
            observation = full_obs

        # Find the healpixel centers that are included in an observation
        indx = self.pointing2hpindx(
            observation["RA"][0], observation["dec"][0], rotSkyPos=observation["rotSkyPos"][0]
        )
        for surveys in self.survey_lists:
            for survey in surveys:
                survey.add_observation(observation, indx=indx)

    def update_conditions(self, conditions_in):
        """
        Parameters
        ----------
        conditions : dict-like
            The current conditions of the telescope (pointing position,
            loaded filters, cloud-mask, etc)
        """
        # Add the current queue and scheduled queue to the conditions
        self.conditions = conditions_in
        # put the local queue in the conditions
        self.conditions.queue = self.queue

        # Check if any surveys have upcomming scheduled observations.
        # Note that we are accumulating all of the possible scheduled
        # observations, so it's up to the user to make sure things don't
        # collide. The ideal implementation would be to have all the
        # scheduled observations in a single survey objects,
        # presumably at the highest tier of priority.

        all_scheduled = []
        for sl in self.survey_lists:
            for sur in sl:
                scheduled = sur.get_scheduled_obs()
                if scheduled is not None:
                    all_scheduled.append(scheduled)
        if len(all_scheduled) == 0:
            self.conditions.scheduled_observations = []
        else:
            all_scheduled = np.sort(np.concatenate(all_scheduled).ravel())
            # In case the surveys have not been removing executed observations
            all_scheduled = all_scheduled[np.where(all_scheduled >= self.conditions.mjd)]
            self.conditions.scheduled_observations = all_scheduled

    def _check_queue_mjd_only(self, mjd):
        """
        Check if there are things in the queue that can be executed
        using only MJD and not full conditions.
        This is primarily used by sim_runner to reduce calls calculating
        updated conditions when they are not needed.
        """
        result = False
        if len(self.queue) > 0:
            if (IntRounded(mjd) < IntRounded(self.queue[0]["flush_by_mjd"])) | (
                self.queue[0]["flush_by_mjd"] == 0
            ):
                result = True
        return result

    def request_observation(self, mjd=None):
        """
        Ask the scheduler what it wants to observe next

        Parameters
        ----------
        mjd : `float` (None)
            The Modified Julian Date.
            If None, it uses the MJD from the conditions from the
            last conditions update.

        Returns
        -------
        observation :  observation object (ra,dec,filter,rotangle)
            Returns None if the queue fails to fill

        Notes
        -----
        Calling request_observation repeatedly, even without updating
        the time or conditions, can return different requested
        observations. This is because the internal queue is filled
        when it becomes empty, and then subsequent calls to
        request_observation will pop successive visits from this queue,
        until the queue is empty or is flushed.
        """
        if mjd is None:
            mjd = self.conditions.mjd
        if len(self.queue) == 0:
            self._fill_queue()

        if len(self.queue) == 0:
            return None
        else:
            # If the queue has gone stale, flush and refill.
            # Zero means no flush_by was set.
            if (IntRounded(mjd) > IntRounded(self.queue[0]["flush_by_mjd"])) & (
                self.queue[0]["flush_by_mjd"] != 0
            ):
                self.flushed += len(self.queue)
                self.flush_queue()
                self._fill_queue()
            if len(self.queue) == 0:
                return None
            observation = self.queue.pop(0)
            # If we are limiting the camera rotator
            if self.rotator_limits is not None:
                alt, az = _approx_ra_dec2_alt_az(
                    observation["RA"],
                    observation["dec"],
                    self.conditions.site.latitude_rad,
                    self.conditions.site.longitude_rad,
                    mjd,
                )
                obs_pa = _approx_altaz2pa(alt, az, self.conditions.site.latitude_rad)
                rot_tel_pos_expected = self.rc._rotskypos2rottelpos(observation["rotSkyPos"], obs_pa)
                if np.isfinite(observation["rotSkyPos"]):
                    if (IntRounded(rot_tel_pos_expected) > IntRounded(self.rotator_limits[0])) & (
                        IntRounded(rot_tel_pos_expected) < IntRounded(self.rotator_limits[1])
                    ):
                        diff = np.abs(self.rotator_limits - rot_tel_pos_expected)
                        limit_indx = np.min(np.where(diff == np.min(diff))[0])
                        observation["rotSkyPos"] = self.rc.rottelpos2rotskypos(
                            self.rotator_limits[limit_indx], obs_pa
                        )
            return observation

    def _fill_queue(self):
        """
        Compute reward function for each survey and fill the observing queue
        with the observations from the highest reward survey.
        """

        # If we loaded the scheduler from a pickle, self.keep_rewards
        # might not have been initialized, but we want it to succeed
        # anyway. So, assign it to false if it isn't present.
        try:
            keep_rewards = self.keep_rewards
        except AttributeError:
            keep_rewards = False

        if keep_rewards:
            # Use perf_counter_ns to get the best time resolution to guarantee
            # successive calls do occur within the same resolution element.
            self.queue_fill_mjd_ns = np.int64(self.mjd_perf_counter_offset + time.perf_counter_ns())
            self.queue_reward_df = self.make_reward_df(accum=False)
            self.queue_reward_df = self.queue_reward_df.assign(
                queue_start_mjd=float(self.conditions.mjd),
                queue_fill_mjd_ns=np.int64(self.queue_fill_mjd_ns),
            )

        rewards = None
        for ns, surveys in enumerate(self.survey_lists):
            rewards = np.zeros(len(surveys))
            for i, survey in enumerate(surveys):
                # For each survey, find the highest reward value.
                rewards[i] = np.nanmax(survey.calc_reward_function(self.conditions))
            # If we have a tier with a good reward, break out of the loop
            if np.nanmax(rewards) > -np.inf:
                self.survey_index[0] = ns
                break
        if (np.nanmax(rewards) == -np.inf) | (np.isnan(np.nanmax(rewards))):
            self.flush_queue()
        else:
            to_fix = np.where(np.isnan(rewards) == True)
            rewards[to_fix] = -np.inf
            # Take a min here, so the surveys will be executed in the order
            # they are entered if there is a tie.
            self.survey_index[1] = np.min(np.where(rewards == np.nanmax(rewards)))
            # Survey return list of observations
            result = self.survey_lists[self.survey_index[0]][self.survey_index[1]].generate_observations(
                self.conditions
            )

            self.queue = result
            self.queue_filled = self.conditions.mjd

        if len(self.queue) == 0:
            self.log.warning(f"Failed to fill queue at time {self.conditions.mjd}")

    def get_basis_functions(self, survey_index=None, conditions=None):
        """Get the basis functions for a specific survey,
        in provided conditions.

        Parameters
        ----------
        survey_index : `List` [`int`]
            A list with two elements: the survey list and the element
            within that survey list for which the basis function should be
            retrieved. If ``None``, use the latest survey to make an addition
            to the queue.
        conditions : `rubin_scheduler.scheduler.features.conditions.Conditions`
            The conditions for which to return the basis functions.
            If ``None``, use the conditions associated with this scheduler.

        Returns
        -------
        basis_funcs : `OrderedDict` ['str`,
        `~.scheduler.basis_functions.basis_functions.Base_basis_function`]
            A dictionary of the basis functions, where the keys are names for
            the basis functions and the values are the functions themselves.
        """

        if survey_index is None:
            survey_index = self.survey_index

        if conditions is None:
            conditions = self.conditions

        survey = self.survey_lists[survey_index[0]][survey_index[1]]
        basis_funcs = OrderedDict()
        if hasattr(survey, "basis_functions"):
            for basis_func in survey.basis_functions:
                sample_values = basis_func(conditions)
                if hasattr(sample_values, "__len__"):
                    key = f"{basis_func.__class__.__name__} @{id(basis_func)}"
                    basis_funcs[key] = basis_func
        return basis_funcs

    def get_healpix_maps(self, survey_index=None, conditions=None):
        """Get the healpix maps for a specific survey, in provided conditions.

        Parameters
        ----------
        survey_index : `List` [`int`], opt
            A list with two elements: the survey list and the element within
            that survey list for which the maps that should be retrieved.
            If ``None``, use the latest survey which added to the queue.
        conditions : `~.scheduler.features.conditions.Conditions`, opt
            The conditions for the maps to be returned. If ``None``, use
            the conditions associated with this sceduler. By default None.

        Returns
        -------
        basis_funcs : `OrderedDict` ['str`, `numpy.ndarray`]
            A dictionary of the maps, where the keys are names for the maps
            and values are the numpy arrays as used by ``healpy``.
        """

        if survey_index is None:
            survey_index = self.survey_index

        if conditions is None:
            conditions = self.conditions

        maps = OrderedDict()
        for band in conditions.skybrightness.keys():
            maps[f"{band}_sky"] = deepcopy(conditions.skybrightness[band])
            maps[f"{band}_sky"][maps[f"{band}_sky"] < -1e30] = np.nan

        basis_functions = self.get_basis_functions(survey_index, conditions)

        for this_basis_func in basis_functions.values():
            label = this_basis_func.label()
            if label in maps:
                label = f"{label} @{id(this_basis_func)}"
            maps[label] = this_basis_func(conditions)

        return maps

    def __repr__(self):
        if isinstance(self.pointing2hpindx, HpInLsstFov):
            camera = "LSST"
        elif isinstance(self.pointing2hpindx, HpInComcamFov):
            camera = "comcam"
        else:
            camera = None

        this_repr = f"""{self.__class__.__qualname__}(
            surveys={repr(self.survey_lists)},
            camera="{camera}",
            nside={repr(self.nside)},
            rotator_limits={repr(self.rotator_limits)},
            survey_index={repr(self.survey_index)},
            log={repr(self.log)}
        )"""
        return this_repr

    def __str__(self):
        # If dependencies of to_markdown are not installed, fall back on repr
        try:
            pd.DataFrame().to_markdown()
        except ImportError:
            return repr(self)

        if isinstance(self.pointing2hpindx, HpInLsstFov):
            camera = "LSST"
        elif isinstance(self.pointing2hpindx, HpInComcamFov):
            camera = "comcam"
        else:
            camera = None

        output = StringIO()
        print(f"# {self.__class__.__name__} at {hex(id(self))}", file=output)

        try:
            last_chosen = str(self.survey_lists[self.survey_index[0]][self.survey_index[1]])
        except TypeError:
            last_chosen = "None"

        misc = pd.Series(
            {
                "camera": camera,
                "nside": self.nside,
                "rotator limits": list(self.rotator_limits),
                "survey index": list(self.survey_index),
                "Last chosen": last_chosen,
            }
        )
        misc.name = "value"
        print(misc.to_markdown(), file=output)

        print("", file=output)
        print("## Surveys", file=output)

        if len(self.survey_lists) == 0:
            print("Scheduler contains no surveys.", file=output)

        for tier_index, tier_surveys in enumerate(self.survey_lists):
            print(file=output)
            print(f"### Survey list {tier_index}", file=output)
            print(self.surveys_df(tier_index).to_markdown(), file=output)

        print("", file=output)
        if hasattr(self, "conditions"):
            print(str(self.conditions), file=output)
        else:
            print("No conditions set", file=output)

        print("", file=output)
        print("## Queue", file=output)
        if len(self.queue) > 0:
            print(
                pd.concat(pd.DataFrame(q) for q in self.queue)[
                    ["ID", "flush_by_mjd", "RA", "dec", "filter", "exptime", "note"]
                ]
                .set_index("ID")
                .to_markdown(),
                file=output,
            )
        else:
            print("Queue is empty", file=output)

        result = output.getvalue()
        return result

    def _repr_markdown_(self):
        # This is used by jupyter
        return str(self)

    def surveys_df(self, tier):
        """Create a pandas.DataFrame describing rewards from surveys.

        Parameters
        ----------
        conditions : `rubin_scheduler.scheduler.features.Conditions`
            Conditions for which rewards are to be returned.
        tier : `int`
            The level of the list of survey lists for which to return values.

        Returns
        -------
        reward_df : `pandas.DataFrame`
            A table of surveys listing the rewards.
        """

        surveys = []
        survey_list = self.survey_lists[tier]
        for survey_list_elem, survey in enumerate(survey_list):
            if (self.survey_index[0] is None) or (tier > self.survey_index[0]):
                # This survey reward was not been evaluated
                reward = None
            elif not hasattr(survey, "reward"):
                reward = None
            elif survey.reward is None:
                reward = None
            elif np.isscalar(survey.reward):
                reward = survey.reward
            elif np.count_nonzero(survey.reward > -np.inf) == 0:
                # The entire survey is masked
                reward = -np.inf
            else:
                reward = np.nanmax(survey.reward)

            try:
                chosen = (tier == self.survey_index[0]) and (survey_list_elem == self.survey_index[1])
            except TypeError:
                chosen = False

            surveys.append({"survey": str(survey), "reward": reward, "chosen": chosen})

        df = pd.DataFrame(surveys).set_index("survey")
        return df

    def make_reward_df(self, conditions=None, accum=True):
        """Create a pandas.DataFrame describing rewards from contained surveys.

        Parameters
        ----------
        conditions : `rubin_scheduler.scheduler.features.Conditions`
            Conditions for which rewards are to be returned
        accum : `bool`
            Include accumulated rewards (defaults to True)

        Returns
        -------
        reward_df : `pandas.DataFrame`
            A table of surveys listing the rewards.
        """
        if conditions is None:
            conditions = self.conditions

        survey_dfs = []
        survey_labels = self.survey_labels
        for index0, survey_list in enumerate(self.survey_lists):
            for index1, survey in enumerate(survey_list):
                survey_df = survey.make_reward_df(conditions, accum=accum)
                if len(survey_df) == 0:
                    continue

                survey_df["list_index"] = index0
                survey_df["survey_index"] = index1
                survey_df["tier_label"] = f"tier {index0}"
                survey_df["survey_label"] = survey_labels[index0][index1]
                survey_df["survey_class"] = survey.__class__.__name__
                survey_df["survey_reward"] = np.nanmax(survey.calc_reward_function(conditions))
                survey_dfs.append(survey_df)

        reward_df = pd.concat(survey_dfs).set_index(["list_index", "survey_index"])
        return reward_df

    @property
    def survey_labels(self):
        """Provide a list of labels for surveys

        Returns
        -------
        label_lists : `list`
            A list, or list of lists, of labels corresponding to
            self.survey_lists
        """
        label_lists = []

        encountered_labels = set()
        duplicated_labels = set()

        def process_survey(survey):
            try:
                basic_label = survey.survey_name
            except AttributeError:
                basic_label = ""

            if len(basic_label) == 0:
                basic_label = survey.__class__.__name__

            if basic_label in encountered_labels:
                duplicated_labels.add(basic_label)
            else:
                encountered_labels.add(basic_label)

            return basic_label

        for item in self.survey_lists:
            if hasattr(item, "generate_observations"):
                basic_label = process_survey(item)
                label_lists.append()
            else:
                labels = []
                for survey in item:
                    basic_label = process_survey(survey)
                    labels.append(basic_label)
                label_lists.append(labels)

        label_count = {}

        def unique_label(survey):
            try:
                basic_label = survey.survey_name
            except AttributeError:
                basic_label = ""

            if len(basic_label) == 0:
                basic_label = survey.__class__.__name__

            if basic_label not in duplicated_labels:
                return basic_label

            if basic_label not in label_count:
                label_count[basic_label] = 0

            label_count[basic_label] += 1
            label = f"{basic_label} {label_count[basic_label]}"
            return label

        label_lists = []
        for item in self.survey_lists:
            if hasattr(item, "generate_observations"):
                label = unique_label(item)
                label_lists.append(label)
            else:
                labels = []
                for survey in item:
                    label = unique_label(survey)
                    labels.append(label)
                label_lists.append(labels)

        return label_lists



__all__ = (
    "_approx_alt_az2_ra_dec",
    "_approx_ra_dec2_alt_az",
    "approx_alt_az2_ra_dec",
    "approx_ra_dec2_alt_az",
    "_approx_altaz2pa",
    "approx_altaz2pa",
)

import numpy as np

from .coordinate_transformations import calc_lmst


def _approx_altaz2pa(alt_rad, az_rad, latitude_rad):
    """A fast calculation of parallactic angle.

    Parameters
    ----------
    alt_rad : `float`
        Altitude (radians)
    az_rad : `float`
        Azimuth (radians)
    latitude_rad : `float`
        The latitude of the observatory (radians)

    Returns
    -------
    pa : `float`
        Parallactic angle (radians)
    """

    y = np.sin(-az_rad) * np.cos(latitude_rad)
    x = np.cos(alt_rad) * np.sin(latitude_rad) - np.sin(alt_rad) * np.cos(latitude_rad) * np.cos(-az_rad)
    pa = np.arctan2(y, x)
    # Make it run from 0-360 deg instead of of -180 to 180
    pa = pa % (2.0 * np.pi)
    return pa


def approx_altaz2pa(alt_deg, az_deg, latitude_deg):
    """A fast calculation of parallactic angle.

    Parameters
    ----------
    alt_rad : `float`
        Altitude (degrees)
    az_rad : `float`
        Azimuth (degrees)
    latitude_rad : `float`
        The latitude of the observatory (degrees)

    Returns
    -------
    pa : `float`
        Parallactic angle (degrees)
    """
    pa = _approx_altaz2pa(np.radians(alt_deg), np.radians(az_deg), np.radians(latitude_deg))
    return np.degrees(pa)


def approx_alt_az2_ra_dec(alt, az, lat, lon, mjd, lmst=None):
    """Convert alt, az to RA, Dec without taking into account aberration,
    precession, diffraction, etc.

    Parameters
    ----------
    alt : numpy.array
        Altitude, same length as `ra` and `dec`. Degrees.
    az : numpy.array
        Azimuth, same length as `ra` and `dec`. Must be same length
        as `alt`. Degrees.
    lat : float
        Latitude of the observatory in degrees.
    lon : float
        Longitude of the observatory in degrees.
    mjd : float
        Modified Julian Date.
    lmst : float (None)
        The local mean sidereal time (computed if not given). (hours)

    Returns
    -------
    ra : array_like
        RA, in degrees.
    dec : array_like
        Dec, in degrees.
    """
    ra, dec = _approx_alt_az2_ra_dec(
        np.radians(alt),
        np.radians(az),
        np.radians(lat),
        np.radians(lon),
        mjd,
        lmst=lmst,
    )
    return np.degrees(ra), np.degrees(dec)


def _approx_alt_az2_ra_dec(alt, az, lat, lon, mjd, lmst=None):
    """Convert alt, az to RA, Dec without taking into account aberration,
    precession, diffraction, etc.

    Parameters
    ----------
    alt : numpy.array
        Altitude, same length as `ra` and `dec`. Radians.
    az : numpy.array
        Azimuth, same length as `ra` and `dec`. Must be same length
        as `alt`. Radians.
    lat : float
        Latitude of the observatory in radians.
    lon : float
        Longitude of the observatory in radians.
    mjd : float
        Modified Julian Date.
    lmst : float (None)
        The local mean sidereal time (computed if not given). (hours)

    Returns
    -------
    ra : array_like
        RA, in radians.
    dec : array_like
        Dec, in radians.
    """
    if lmst is None:
        lmst = calc_lmst(mjd, lon)
    lmst = lmst / 12.0 * np.pi  # convert to rad
    sindec = np.sin(lat) * np.sin(alt) + np.cos(lat) * np.cos(alt) * np.cos(az)
    sindec = np.clip(sindec, -1, 1)
    dec = np.arcsin(sindec)
    ha = np.arctan2(
        -np.sin(az) * np.cos(alt),
        -np.cos(az) * np.sin(lat) * np.cos(alt) + np.sin(alt) * np.cos(lat),
    )
    ra = lmst - ha
    raneg = np.where(ra < 0)
    ra[raneg] = ra[raneg] + 2.0 * np.pi
    raover = np.where(ra > 2.0 * np.pi)
    ra[raover] -= 2.0 * np.pi
    return ra, dec


def approx_ra_dec2_alt_az(ra, dec, lat, lon, mjd, lmst=None):
    """Convert Ra,Dec to Altitude and Azimuth.

    Coordinate transformation is killing performance. Just use
    simple equations to speed it up and ignore aberration, precession,
    nutation, nutrition, etc.

    Parameters
    ----------
    ra : array_like
        RA, in degrees.
    dec : array_like
        Dec, in degrees. Must be same length as `ra`.
    lat : float
        Latitude of the observatory in degrees.
    lon : float
        Longitude of the observatory in degrees.
    mjd : float
        Modified Julian Date.
    lmst : float (None)
        The local mean sidereal time (computed if not given). (hours)

    Returns
    -------
    alt : numpy.array
        Altitude, same length as `ra` and `dec`. degrees.
    az : numpy.array
        Azimuth, same length as `ra` and `dec`. degrees.
    """
    alt, az = _approx_ra_dec2_alt_az(
        np.radians(ra),
        np.radians(dec),
        np.radians(lat),
        np.radians(lon),
        mjd,
        lmst=lmst,
    )
    return np.degrees(alt), np.degrees(az)


def _approx_ra_dec2_alt_az(ra, dec, lat, lon, mjd, lmst=None, return_pa=False):
    """Convert Ra,Dec to Altitude and Azimuth.

    Coordinate transformation is killing performance. Just use
    simple equations to speed it up and ignore aberration,
    precession, nutation, nutrition, etc.

    Parameters
    ----------
    ra : array_like
        RA, in radians.
    dec : array_like
        Dec, in radians. Must be same length as `ra`.
    lat : float
        Latitude of the observatory in radians.
    lon : float
        Longitude of the observatory in radians.
    mjd : float
        Modified Julian Date.
    lmst : float (None)
        The local mean sidereal time (computed if not given). (hours)

    Returns
    -------
    alt : numpy.array
        Altitude, same length as `ra` and `dec`. Radians.
    az : numpy.array
        Azimuth, same length as `ra` and `dec`. Radians.
    """
    if lmst is None:
        lmst = calc_lmst(mjd, lon)
    lmst = lmst / 12.0 * np.pi  # convert to rad
    ha = lmst - ra
    sindec = np.sin(dec)
    sinlat = np.sin(lat)
    coslat = np.cos(lat)
    sinalt = sindec * sinlat + np.cos(dec) * coslat * np.cos(ha)
    sinalt = np.clip(sinalt, -1, 1)
    alt = np.arcsin(sinalt)
    cosaz = (sindec - np.sin(alt) * sinlat) / (np.cos(alt) * coslat)
    cosaz = np.clip(cosaz, -1, 1)
    az = np.arccos(cosaz)
    if np.size(ha) < 2:
        if np.sin(ha) > 0:
            az = 2.0 * np.pi - az
    else:
        signflip = np.where(np.sin(ha) > 0)
        az[signflip] = 2.0 * np.pi - az[signflip]
    if return_pa:
        pa = _approx_altaz2pa(alt, az, lat)
        return alt, az, pa
    return alt, az
